import pandas as pd
import numpy as np

import jax
from jax import numpy as jnp

from functools import partial
from sklearn import preprocessing

from .utils import (
    dict_to_binary,
    get_expr,
    frac_nonzero,
    masked_max,
    masked_mean,
    jit_frac_nonzero,
)


def annotate_snap(
    adata,
    marker_dict,
    group_name,
    layer=None,
    auc_weight=0.5,
    expr_weight=0.5,
    marker_summary_fun="max",
    apply_fun="vmap",
):
    """
    Annotate cell types based on AUROC and expression of predefined marker genes.

    Args:
        adata: AnnData object
        marker_dict: Dict with marker genes for each celltype
        group_name: Name of the column in adata.obs that contains the cluster labels
        layer: Layer in adata to use for expression
        auc_weight: Weight to give to AUROC in the final score
        expr_weight: Weight to give to detection rate in the final score
        marker_summary_fun: Function to use to summarize over markers for the same
            annotation. Options are "max" and "mean".
        apply_fun: Function to use for applying the auroc function. Options are "vmap"
            and "numpy". vmap is a lot faster but can cause OOM errors for big matrices.
    """
    # Reformat marker_dict into binary matrix
    marker_mat = dict_to_binary(marker_dict)

    # Compute AUROC and fraction nonzero for marker features
    features = marker_mat.columns
    metrics = auc_expr(
        adata, group_name, features=features, layer=layer, apply_fun=apply_fun
    )

    marker_mat = marker_mat.loc[:, metrics["features"]]
    if marker_summary_fun == "max":
        auc_max = masked_max(metrics["auroc"], marker_mat.values)
        expr_max = masked_max(metrics["frac_nonzero"], marker_mat.values)
    elif marker_summary_fun == "mean":
        auc_max = masked_mean(metrics["auroc"], marker_mat.values)
        expr_max = masked_mean(metrics["frac_nonzero"], marker_mat.values)
    else:
        raise ValueError(
            f"Invalid marker_summary_fun: {marker_summary_fun}. "
            "Current options are 'max' and 'mean'."
        )
    # Combine metrics
    assignment_scores = (auc_weight * auc_max + expr_weight * expr_max) / (
        auc_weight + expr_weight
    )
    assign_idx = jnp.argmax(assignment_scores, axis=0)
    # Mask out genes that are not expressed in any cell
    assign_class = marker_mat.index[assign_idx]

    assign_df = pd.DataFrame(
        {
            "class": assign_class,
            "score": assignment_scores[assign_idx, jnp.arange(auc_max.shape[1])],
            "auc": auc_max[assign_idx, jnp.arange(auc_max.shape[1])],
            "expr": expr_max[assign_idx, jnp.arange(expr_max.shape[1])],
        },
        index=metrics["groups"],
    )

    return assign_df


def auc_expr(
    adata,
    group_name,
    features=None,
    layer=None,
    compute_auroc=True,
    compute_frac_nonzero=True,
    compute_frac_nonzero_out=False,
    apply_fun="vmap",
):
    """Computes AUROC and fraction nonzero for each gene in an adata object."""
    # Turn string groups into integers
    le = preprocessing.LabelEncoder()
    le.fit(adata.obs[group_name])

    # Compute AUROC and fraction nonzero
    groups = jnp.array(le.transform(adata.obs[group_name]))
    expr, features = get_expr(adata, features=features, layer=layer)
    auroc, frac_nonzero, frac_nonzero_out = expr_auroc_over_groups(
        expr,
        groups,
        compute_auroc=compute_auroc,
        compute_frac_nz=compute_frac_nonzero,
        compute_frac_nz_out=compute_frac_nonzero_out,
        apply_fun=apply_fun,
    )

    return dict(
        frac_nonzero=frac_nonzero,
        frac_nonzero_out=frac_nonzero_out,
        auroc=auroc,
        features=features,
        groups=le.classes_,
    )


@jax.jit
def jit_auroc(x, groups):
    # TODO: compute frac nonzero here to avoid iterating twice

    # sort scores and corresponding truth values
    desc_score_indices = jnp.argsort(x)[::-1]
    x = x[desc_score_indices]
    groups = groups[desc_score_indices]

    # x typically has many tied values. Here we extract
    # the indices associated with the distinct values. We also
    # concatenate a value for the end of the curve.
    distinct_value_indices = jnp.array(jnp.diff(x) != 0, dtype=jnp.int32)
    threshold_mask = jnp.r_[distinct_value_indices, 1]

    # accumulate the true positives with decreasing threshold
    tps_ = jnp.cumsum(groups)
    fps_ = 1 + jnp.arange(groups.size) - tps_

    # mask out the values that are not distinct
    tps = jnp.sort(tps_ * threshold_mask)
    fps = jnp.sort(fps_ * threshold_mask)
    tps = jnp.r_[0, tps]
    fps = jnp.r_[0, fps]
    fpr = fps / fps[-1]
    tpr = tps / tps[-1]
    area = jnp.trapezoid(tpr, fpr)
    return area


vmap_auroc = jax.vmap(jit_auroc, in_axes=[1, None])


def numpy_auroc(x, groups):
    jit_auroc_ = partial(jit_auroc, groups=groups)
    auroc = np.apply_along_axis(jit_auroc_, axis=0, arr=x)
    return auroc


def expr_auroc_over_groups(
    expr,
    groups,
    compute_auroc=True,
    compute_frac_nz=True,
    compute_frac_nz_out=False,
    apply_fun="vmap",
):
    """Computes AUROC for each group separately."""
    auroc = np.zeros((groups.max() + 1, expr.shape[1]))
    frac_nz = np.zeros((groups.max() + 1, expr.shape[1]))
    frac_nz_out = np.zeros((groups.max() + 1, expr.shape[1]))

    if apply_fun == "vmap":
        auroc_fun = vmap_auroc
        fracnz_fun = jit_frac_nonzero
    elif apply_fun == "numpy":
        auroc_fun = numpy_auroc
        fracnz_fun = frac_nonzero

    for group in range(groups.max() + 1):
        if compute_auroc:
            auroc[group, :] = np.array(auroc_fun(expr, groups == group))
        if compute_frac_nz:
            frac_nz[group, :] = np.array(fracnz_fun(expr[groups == group, :]))
        if compute_frac_nz_out:
            frac_nz_out[group, :] = np.array(fracnz_fun(expr[groups != group, :]))

    return auroc, frac_nz, frac_nz_out



import yaml
import pandas as pd
import numba

import jax
from jax import numpy as jnp
from jax.experimental.sparse import BCOO

from functools import partial
from sklearn.metrics import roc_auc_score

import scanpy as sc


def to_jax_array(x):
    """Turn matrix to jax array."""
    if hasattr(x, "todense"):
        # Currently this is not supported for really large matrices
        # return BCOO.from_scipy_sparse(x).update_layout(n_batch=1, on_inefficient=None)
        return jnp.asarray(x.todense())
    else:
        return jnp.asarray(x)


@jax.jit
@partial(jax.vmap, in_axes=[None, 0])
def masked_max(x, mask):
    return jnp.max(x * mask, axis=1)


@jax.jit
@partial(jax.vmap, in_axes=[None, 0])
def masked_mean(x, mask):
    return jnp.sum(x * mask, axis=1) / jnp.sum(mask)


def frac_nonzero(x, axis=0):
    return jnp.mean(x > 0, axis=axis)


jit_frac_nonzero = jax.jit(frac_nonzero)


def dict_to_binary(d):
    df = pd.concat(
        [pd.Series(v, name=k).astype(str) for k, v in d.items()],
        axis=1,
    )
    marker_mat = pd.get_dummies(df.stack()).groupby(level=1).sum().clip(upper=1)
    return marker_mat


@numba.jit
def match(a, b):
    return [b.index(x) if x in b else None for x in a]


def get_expr(adata, features=None, layer=None):
    """Get expression matrix from adata object"""
    if features is not None:
        # intersect with adata features
        features = list(set(features) & set(adata.var_names))
        adata = adata[
            :,
            match(
                numba.typed.List(features), numba.typed.List(adata.var_names.tolist())
            ),
        ]
    else:
        features = adata.var_names.copy().tolist()

    if layer is not None:
        expr = to_jax_array(adata.layers[layer])
    else:
        expr = to_jax_array(adata.X)

    return expr, features


def get_markers(x):
    return {n: v["marker_genes"] for n, v in x.items()}


def read_yaml(file):
    with open(file, "r") as f:
        marker_dict = yaml.safe_load(f)
    return marker_dict


def get_annot_df(x, group_name, min_expr=0.1):
    # Get valid annots from each level
    annot_list = []
    for k, v in x.items():
        annot = v.set_index(group_name)["class"]
        if min_expr > 0:
            expr = v.set_index(group_name)["expr"]
            annot = annot[expr > min_expr]
        annot_list.append(annot)
    # Concat annots
    annot_df = pd.concat(annot_list, axis=1)
    # Rename cols to levels
    annot_df.columns = [str(i) for i in x.keys()]
    return annot_df


def matrix_to_long_df(x, features, groups):
    """Converts a matrix to a long dataframe"""
    df = pd.DataFrame(x, index=groups, columns=features)
    df = df.stack().reset_index()
    df.columns = ["group", "feature", "value"]
    return df



from typing import Optional
import anndata as ad
import pandas as pd

from .auroc import annotate_snap

from .utils import get_markers, get_annot_df


def annotate_hierarchy(
    adata: ad.AnnData,
    marker_hierarchy: dict,
    group_name: str,
    layer: Optional[str] = None,
    min_expr: float = 0.1,
    **kwargs
) -> dict:
    """
    Annotate clusters based on a manually defined cell type and marker hierarchy.

    Args:
        adata: AnnData object
        marker_hierarchy: dict arker genes for each celltype arranged
            hierarchically.
        group_name: Name of the column in adata.obs that contains the cluster labels
        layer: Layer in adata to use for expression
        **kwargs: Additional arguments to pass to the annotation function.

    Returns:
        Dict with assignments and metrics
    """

    # Annotate at each level of the hierarchy
    assignment_hierarchy = annotate_levels(
        adata, marker_hierarchy, group_name, **kwargs
    )

    return dict(
        assignments=get_annot_df(assignment_hierarchy, group_name, min_expr=min_expr),
        metrics=assignment_hierarchy,
    )


def annotate_levels(
    adata,
    marker_hierarchy,
    group_name,
    level=0,
    assignment_levels=None,
    layer=None,
    **kwargs
):
    """Recursively annotatates all levels of a marker hierarchy."""
    level += 1
    level_name = "level_" + str(level)
    marker_dict = get_markers(marker_hierarchy)
    assignments = annotate(adata, marker_dict, group_name, layer=layer, **kwargs)

    if assignment_levels is None:
        assignment_levels = {}

    if level_name not in assignment_levels.keys():
        assignment_levels[level_name] = pd.DataFrame()

    assignment_levels[level_name] = pd.concat(
        [assignment_levels[level_name], assignments], axis=0
    )

    for subtype in assignments["class"].unique():

        if "subtypes" not in marker_hierarchy[subtype].keys():
            continue

        # Subset adata
        subtype_groups = assignments[group_name][
            assignments["class"] == subtype
        ].astype(str)
        subtype_adata = adata[adata.obs[group_name].isin(subtype_groups)]

        # Recursively annotate
        assignment_levels = annotate_levels(
            subtype_adata,
            marker_hierarchy[subtype]["subtypes"],
            group_name,
            level=level,
            assignment_levels=assignment_levels,
            layer=layer,
        )

    return assignment_levels


def annotate(
    adata: ad.AnnData,
    marker_dict: dict,
    group_name: str,
    layer: Optional[str] = None,
    **kwargs
) -> pd.DataFrame:
    """
    Annotate clusters based on a manually defined cell type markers.

    Args:
        adata: AnnData object
        marker_dict: Dict with marker genes for each celltype
        group_name: Name of the column in adata.obs that contains the cluster labels
        layer: Layer in adata to use for expression
        **kwargs: Additional arguments to pass to the annotation function.

    Returns:
        pd.DataFrame with assignments
    """
    assignments = annotate_snap(adata, marker_dict, group_name, layer=layer, **kwargs)
    # Join cluster-level results with adata
    assignments = assignments.reset_index(names=group_name)
    return assignments



import anndata as ad
import pandas as pd
import numpy as np
import scanpy as sc


def compute_glycolysis_score(adata: ad.AnnData, **kwargs):
    """Compute glycolysis score using hallmark glycolysis genes from MSigDB.

    Args:
        adata: Anndata object with gene expression data.
        **kwargs: Additional arguments passed to `sc.tl.score_genes`.

    Returns:
        `None`. The glycolysis score is stored in `adata.obs["Hallmark_Glycolysis"]`.
    """
    msigdb_glycolysis = np.array(
        pd.read_csv(
            "https://www.gsea-msigdb.org/gsea/msigdb/human/download_geneset.jsp?geneSetName=HALLMARK_GLYCOLYSIS&fileType=TSV",
            sep="\t",
            header=None,
            index_col=0,
        )
        .loc["GENE_SYMBOLS", 1]
        .split(",")
    )
    msigdb_glycolysis = np.intersect1d(msigdb_glycolysis, adata.var_names)
    sc.tl.score_genes(
        adata, msigdb_glycolysis, score_name="Hallmark_Glycolysis", **kwargs
    )



import scanpy as sc
import numpy as np
import pandas as pd
import anndata as ad
import torch
from pynndescent import NNDescent

from scipy import sparse
from typing import Optional, Union, Mapping, Literal
import warnings
import sys
import os
import importlib.util
import argparse
import tqdm

warnings.filterwarnings("ignore")


def nn2adj_gpu(nn, n1=None, n2=None):
    if n1 is None:
        n1 = nn[1].shape[0]
    if n2 is None:
        n2 = np.max(nn[1].flatten())

    df = pd.DataFrame(
        {
            "i": np.repeat(range(nn[1].shape[0]), nn[1].shape[1]),
            "j": nn[1].flatten(),
            "x": nn[0].flatten(),
        }
    )
    adj = sparse.csr_matrix(
        (np.repeat(1, df.shape[0]), (df["i"], df["j"])), shape=(n1, n2)
    )

    return adj


def nn2adj_cpu(nn, n1=None, n2=None):
    if n1 is None:
        n1 = nn[0].shape[0]
    if n2 is None:
        n2 = np.max(nn[0].flatten())

    df = pd.DataFrame(
        {
            "i": np.repeat(range(nn[0].shape[0]), nn[0].shape[1]),
            "j": nn[0].flatten(),
            "x": nn[1].flatten(),
        }
    )

    adj = sparse.csr_matrix(
        (np.repeat(1, df.shape[0]), (df["i"], df["j"])), shape=(n1, n2)
    )

    return adj


def build_nn(
    ref,
    query=None,
    k=100,
    weight: Literal["unweighted", "dist", "gaussian_kernel"] = "unweighted",
    sigma=None,
):
    if query is None:
        query = ref

    if torch.cuda.is_available() and importlib.util.find_spec("cuml"):
        print(
            "Info: GPU detected and cuml installed. Use cuML for neighborhood estimation."
        )
        from cuml.neighbors import NearestNeighbors

        model = NearestNeighbors(n_neighbors=k)
        model.fit(ref)
        knn = model.kneighbors(query)
        adj = nn2adj_gpu(knn, n1=query.shape[0], n2=ref.shape[0])
    else:
        print(
            "Info: Failed calling cuML. Falling back to neighborhood estimation using CPU with pynndescent."
        )
        index = NNDescent(ref)
        knn = index.query(query, k=k)
        adj = nn2adj_cpu(knn, n1=query.shape[0], n2=ref.shape[0])

    return adj


def build_mutual_nn(dat1, dat2=None, k1=100, k2=None):
    if dat2 is None:
        dat2 = dat1
    if k2 is None:
        k2 = k1

    adj_12 = build_nn(dat1, dat2, k=k2)
    adj_21 = build_nn(dat2, dat1, k=k1)

    adj_mnn = adj_12.multiply(adj_21.T)
    return adj_mnn


def get_transition_prob_mat(dat, k=50, symm=True):
    adj = build_nn(dat, k=k)
    if symm:
        adj = adj + adj.transpose()
    prob = sparse.diags(1 / np.array(adj.sum(1)).flatten()) @ adj.transpose()
    return prob


def random_walk_with_restart(init, transition_prob, alpha=0.5, num_rounds=100):
    init = np.array(init).flatten()
    heat = init[:, None]
    for i in range(num_rounds):
        heat = init[:, None] * alpha + (1 - alpha) * (
            transition_prob.transpose() @ heat
        )
    return heat


def get_wknn(
    ref,
    query,
    ref2=None,
    k: int = 100,
    query2ref: bool = True,
    ref2query: bool = False,
    weighting_scheme: Literal[
        "n", "top_n", "jaccard", "jaccard_square", "gaussian", "dist"
    ] = "jaccard_square",
    top_n: Optional[int] = None,
    return_adjs: bool = False,
):
    """
    Compute the weighted k-nearest neighbors graph between the reference and query datasets

    Parameters
    ----------
    ref : np.ndarray
        The reference representation to build ref-query neighbor graph
    query : np.ndarray
        The query representation to build ref-query neighbor graph
    ref2 : np.ndarray
        The reference representation to build ref-ref neighbor graph
    k : int
        Number of neighbors per cell
    query2ref : bool
        Consider query-to-ref neighbors
    ref2query : bool
        Consider ref-to-query neighbors
    weighting_scheme : str
        How to weight edges in the ref-query neighbor graph
    top_n : int
        The number of top neighbors to consider
    return_adjs : bool
        Whether to return the adjacency matrices of ref-query, query-ref, ref-ref, and ref-ref for weighting
    """
    adj_q2r = build_nn(ref=ref, query=query, k=k)

    adj_r2q = None
    if ref2query:
        adj_r2q = build_nn(ref=query, query=ref, k=k)

    if query2ref and not ref2query:
        adj_knn = adj_q2r.T
    elif ref2query and not query2ref:
        adj_knn = adj_r2q
    elif ref2query and query2ref:
        adj_knn = ((adj_r2q + adj_q2r.T) > 0) + 0
    else:
        warnings.warn(
            "At least one of query2ref and ref2query should be True. Reset to default with both being True."
        )
        adj_knn = ((adj_r2q + adj_q2r.T) > 0) + 0

    if ref2 is None:
        ref2 = ref
    adj_ref = build_nn(ref=ref2, k=k)
    print("Info: Computing shared neighbors")
    num_shared_neighbors = adj_q2r @ adj_ref.T
    print("Info: Computing weighted neighbors")
    num_shared_neighbors_nn = num_shared_neighbors.multiply(adj_knn.T)

    wknn = num_shared_neighbors_nn.copy()
    if weighting_scheme == "top_n":
        if top_n is None:
            top_n = k // 4 if k > 4 else 1
        wknn = (wknn > top_n) * 1
    elif weighting_scheme == "jaccard":
        wknn.data = wknn.data / (k + k - wknn.data)
    elif weighting_scheme == "jaccard_square":
        wknn.data = (wknn.data / (k + k - wknn.data)) ** 2

    if return_adjs:
        adjs = {"q2r": adj_q2r, "r2q": adj_r2q, "knn": adj_knn, "r2r": adj_ref}
        return (wknn, adjs)
    else:
        return wknn


def estimate_presence_score(
    ref_adata,
    query_adata,
    wknn=None,
    use_rep_ref_wknn="X_latent",
    use_rep_query_wknn="X_latent",
    k_wknn=100,
    query2ref_wknn=True,
    ref2query_wknn=False,
    weighting_scheme_wknn="jaccard_square",
    ref_trans_prop=None,
    use_rep_ref_trans_prop=None,
    k_ref_trans_prop=50,
    symm_ref_trans_prop=True,
    split_by=None,
    do_random_walk=True,
    alpha_random_walk=0.1,
    num_rounds_random_walk=100,
    log=True,
):
    if wknn is None:
        ref = ref_adata.obsm[use_rep_ref_wknn]
        query = query_adata.obsm[use_rep_query_wknn]
        wknn = get_wknn(
            ref=ref,
            query=query,
            k=k_wknn,
            query2ref=query2ref_wknn,
            ref2query=ref2query_wknn,
            weighting_scheme=weighting_scheme_wknn,
        )

    if ref_trans_prop is None and do_random_walk:
        if use_rep_ref_trans_prop is None:
            use_rep_ref_trans_prop = use_rep_ref_wknn
        ref = ref_adata.obsm[use_rep_ref_trans_prop]
        ref_trans_prop = get_transition_prob_mat(ref, k=k_ref_trans_prop)

    if split_by and split_by in query_adata.obs.columns:
        presence_split = [
            np.array(wknn[query_adata.obs[split_by] == x, :].sum(axis=0)).flatten()
            for x in query_adata.obs[split_by].unique()
        ]
    else:
        presence_split = [np.array(wknn.sum(axis=0)).flatten()]

    if do_random_walk:
        print("Info: Smoothing presence scores with random walk")
        presence_split_sm = []
        for x in tqdm.tqdm(presence_split):
            presence_split_sm.append(
                random_walk_with_restart(
                    init=x,
                    transition_prob=ref_trans_prop,
                    alpha=alpha_random_walk,
                    num_rounds=num_rounds_random_walk,
                )
            )
    else:
        presence_split_sm = [x[:, None] for x in presence_split]

    columns = (
        query_adata.obs[split_by].unique()
        if split_by and split_by in query_adata.obs.columns
        else ["query"]
    )
    if len(columns) > 1:
        df_presence = pd.DataFrame(
            np.concatenate(presence_split_sm, axis=1),
            columns=columns,
            index=ref_adata.obs_names,
        )
    else:
        df_presence = pd.DataFrame(
            {columns[0]: presence_split_sm[0].flatten()}
        ).set_index(ref_adata.obs_names)

    if log:
        df_presence = df_presence.apply(lambda x: np.log1p(x), axis=0)
    df_presence_norm = df_presence.apply(
        lambda x: np.clip(x, np.percentile(x, 1), np.percentile(x, 99)), axis=0
    ).apply(lambda x: (x - np.min(x)) / (np.max(x) - np.min(x)), axis=0)
    max_presence = df_presence_norm.max(1)

    return {
        "max": max_presence,
        "per_group": df_presence_norm,
        "ref_trans_prop": ref_trans_prop,
    }


def transfer_labels(ref_adata, query_adata, wknn, label_key="celltype"):
    scores = pd.DataFrame(
        wknn @ pd.get_dummies(ref_adata.obs[label_key]),
        columns=pd.get_dummies(ref_adata.obs[label_key]).columns,
        index=query_adata.obs_names,
    )
    scores["best_label"] = scores.idxmax(1)
    scores["best_score"] = scores.max(1)
    return scores



import numpy as np
import pandas as pd
from scipy.sparse import issparse
from scipy.stats import f
from statsmodels.formula.api import ols
from statsmodels.stats.anova import anova_lm
from joblib import Parallel, delayed, wrap_non_picklable_objects
import multiprocessing
from tqdm import tqdm


def anova(e, group, covar=None, return_coef_group=None):
    if issparse(e):
        e = e.todense()
    e = np.array(e).flatten()

    if covar is None:
        m0 = ols("e ~ 1", data=pd.DataFrame({"e": e})).fit()
        m1 = ols("e ~ zz_group", data=pd.DataFrame({"e": e, "zz_group": group})).fit()
    else:
        data = pd.DataFrame({"e": e, **covar, "zz_group": group})
        m0 = ols("e ~ " + " + ".join(covar.columns), data=data).fit()
        m1 = ols("e ~ " + " + ".join(covar.columns) + " + zz_group", data=data).fit()

    a0 = anova_lm(m0)
    a1 = anova_lm(m1)
    a01 = anova_lm(m0, m1)
    p1 = a01.loc[:, "Pr(>F)"].iloc[-1]
    p2_cdf = f.cdf(
        a0.loc["Residual", "mean_sq"] / a1.loc["Residual", "mean_sq"],
        a0.loc["Residual", "df"],
        a1.loc["Residual", "df"],
    )
    p2 = 1 - np.abs(0.5 - p2_cdf) * 2

    var_tot = a0["sum_sq"].sum()
    var_covar = a0.iloc[:-1]["sum_sq"].sum()
    var_g = a1.loc["zz_group", "sum_sq"]

    if (
        return_coef_group is not None
        and return_coef_group in pd.Series(group).unique()
        and f"zz_group[T.{return_coef_group}]" in m1.params.index
    ):
        coef_g = m1.params[f"zz_group[T.{return_coef_group}]"]
    else:
        coef_g = np.nan

    return (var_g, var_covar, var_tot, p1, p2, coef_g)


delayed_anova = delayed(wrap_non_picklable_objects(anova))


def ancova_group_test(
    expr, group, covar=None, num_threads=1, return_coef_group=None, var_names=None
):
    if not isinstance(group, pd.Series):
        group = pd.Series(group)
    idx = np.where(~group.isnull())[0]
    expr = expr[idx, :]
    group = group[idx]
    if covar is not None:
        covar = pd.DataFrame(covar).iloc[idx, :]

    if group.dtype.name != "category":
        group = group.astype("category")
    group = group.cat.remove_unused_categories()
    g_lev = np.array(pd.Series(group).astype("category").cat.categories)
    if return_coef_group is not None and return_coef_group in g_lev:
        g_lev = np.concatenate([g_lev[g_lev != return_coef_group], [return_coef_group]])
        group = group.cat.reorder_categories(g_lev)

    num_cores = min(num_threads, multiprocessing.cpu_count())
    if num_cores > 1:
        results = Parallel(n_jobs=num_cores)(
            delayed_anova(expr[:, i], group, covar, return_coef_group)
            for i in range(expr.shape[1])
        )
    else:
        results = list()
        for i in tqdm(range(expr.shape[1])):
            results.append(
                anova(
                    expr[:, i], group, covar=covar, return_coef_group=return_coef_group
                )
            )

    df_res = pd.DataFrame(
        results,
        columns=["var_group", "var_covar", "var_total", "p_ANCOVA", "p_Resi", "coef"],
    )
    if var_names is not None and len(var_names) == df_res.shape[0]:
        df_res.index = var_names
    else:
        df_res.index = pd.RangeIndex(start=0, stop=expr.shape[1], step=1)
    return df_res


def f_nonzero(e, covar=None):
    if issparse(e):
        e = e.todense()
    e = np.array(e).flatten()

    if covar is None:
        m1 = ols("e ~ 1", data=pd.DataFrame({"e": e})).fit()
        a1 = anova_lm(m1)
        fval = (np.sum(e**2) / len(e)) / a1.loc["Residual", "mean_sq"]
        p = 1 - np.abs(0.5 - f.cdf(fval, len(e), a1.loc["Residual", "df"])) * 2
        coef = m1.params["Intercept"]
    else:
        data = pd.DataFrame({"e": e, **covar})
        m0 = ols("e ~ " + " + ".join(covar.columns) + " -1", data=data).fit()
        a0 = anova_lm(m0)
        m1 = ols("e ~ " + " + ".join(covar.columns), data=data).fit()
        a1 = anova_lm(m1)
        sumsq0 = a0.loc["Residual", "sum_sq"] + (
            a0.iloc[:-1]["sum_sq"].sum() - a1.iloc[:-1]["sum_sq"].sum()
        )
        meansq0 = sumsq0 / (a0.loc["Residual", "df"] + 1)
        meansq1 = a1.loc["Residual", "mean_sq"]
        fval = meansq0 / meansq1
        p = (
            1
            - np.abs(
                0.5
                - f.cdf(fval, a0.loc["Residual", "df"] + 1, a1.loc["Residual", "df"])
            )
            * 2
        )
        coef = m1.params["Intercept"]

    return (fval, coef, p)


delayed_f_nonzero = delayed(wrap_non_picklable_objects(f_nonzero))


def f_nonzero_test(expr, covar=None, num_threads=1, var_names=None):
    if covar is not None:
        covar = pd.DataFrame(covar)

    num_cores = min(num_threads, multiprocessing.cpu_count())
    if num_cores > 1:
        results = Parallel(n_jobs=num_cores)(
            delayed_f_nonzero(expr[:, i], covar) for i in range(expr.shape[1])
        )
    else:
        results = list()
        for i in tqdm(range(expr.shape[1])):
            results.append(f_nonzero(expr[:, i], covar=covar))

    df_res = pd.DataFrame(results, columns=["f", "coef", "pval"])
    if var_names is not None and len(var_names) == df_res.shape[0]:
        df_res.index = var_names
    else:
        df_res.index = pd.RangeIndex(start=0, stop=expr.shape[1], step=1)
    return df_res



from typing import Optional
import anndata as ad
import numpy as np
import pandas as pd
import scanpy as sc
from decoupler import get_pseudobulk


def create_pseudobulk(
    adata: ad.AnnData,
    sample_key: str = "batch",
    group_key: Optional[str] = None,
    layer: Optional[str] = None,
    mode: str = "sum",
    min_cells: int = 10,
    min_counts: int = 1000,
    **kwargs,
) -> ad.AnnData:
    """
    Create pseudobulk data from anndata object

    Args:
        adata: AnnData object
        sample_key: Column name in adata.obs that contains the sample ID
        group_key: Column name in adata.obs that contains the group ID
        layer: Layer to use for pseudobulk data. If None, use `adata.X`
        mode: Method to aggregate data. Default is 'sum'.
        min_cells: Filter to remove samples by a minimum number of cells in a sample-group pair.
        min_counts: Filter to remove samples by a minimum number of summed counts in a sample-group pair.
        **kwargs: Additional arguments to pass to `decoupler.get_pseudobulk()`


    Returns:
        AnnData object with pseudobulk data
    """
    if layer is not None:
        adata.X = adata.layers[layer].copy()

    adata.obs["n_genes"] = (adata.X > 0).sum(axis=1).A.ravel()

    if group_key is not None:
        adata.obs["group"] = adata.obs[group_key]
    else:
        adata.obs["group"] = "group"

    adata_pb = get_pseudobulk(
        adata,
        sample_col=sample_key,
        groups_col="group",
        layer=layer,
        mode=mode,
        min_cells=min_cells,
        min_counts=min_counts,
        **kwargs,
    )
    adata_pb.X = np.round(adata_pb.X)
    adata_pb.obs["n_genes_median"] = [
        adata.obs.groupby("batch").median()["n_genes"].loc[i.split("_onegroup")[0]]
        for i in adata_pb.obs.index
    ]
    adata_pb.obs["n_genes_std"] = [
        adata.obs.groupby("batch").std()["n_genes"].loc[i.split("_onegroup")[0]]
        for i in adata_pb.obs.index
    ]

    return adata_pb



# Actually just many colors for plotting clusters
# The first then or so are nice, then it starts to get random

many = [
    "#15889C",
    "#ED496F",
    "#8E1382",
    "#FFB300",
    "#FF6F00",
    "#43A047",
    "#C62828",
    "#2d62a3",
    "#1F7530",
    "#573794",
    "#28AAE1",
    "#870f0f",
    "#9e9e9e",
    "#795548",
    "#CD6155",
    "#F7DC6F",
    "#7DCEA0",
    "#85C1E9",
    "#EB984E",
    "#ead9d5",
    "#03cd4A",
    "#CDDC39",
    "#e0594b",
    "#c76cde",
    "#24B177",
    "#8D6E63",
    "#486ff7",
    "#6300b5",
    "#88e200",
    "#012824",
    "#0d3290",
    "#a347fb",
    "#54fc7a",
    "#eb1388",
    "#b0978d",
    "#fe52cf",
    "#83f1f6",
    "#f1f847",
    "#2b1dfc",
    "#6c6f15",
    "#6ca05c",
    "#7788cd",
    "#f502f3",
    "#0dc290",
    "#fa0e03",
    "#3caa0a",
    "#befc8d",
    "#08f8eb",
    "#b1cd3f",
    "#d6a5fa",
    "#ce606c",
    "#ab1eba",
    "#6ecc9f",
    "#054ddc",
    "#486ff7",
    "#854f49",
    "#f22B21",
    "#3a0e43",
    "#225805",
    "#37d160",
    "#e4b974",
    "#a8bade",
    "#47edd1",
    "#f47a92",
    "#c76cde",
    "#9106eb",
    "#81aa20",
    "#d7fdfd",
    "#5deb2e",
    "#f82745",
    "#6435e0",
    "#027ffe",
    "#8e3101",
    "#16f648",
    "#1c15bc",
    "#8be46e",
    "#8d6fa0",
    "#e68fc6",
    "#058ca9",
    "#9e018a",
    "#bdfd0b",
    "#b22760",
    "#2bf49f",
    "#cb9348",
    "#9d8303",
    "#c251a1",
    "#46adaf",
    "#a3e3af",
    "#22bb34",
    "#6ea3fa",
    "#260374",
    "#1c3854",
    "#405d37",
    "#c21df3",
    "#fcea92",
    "#537f88",
    "#fd4c18",
    "#f2d71e",
    "#fd4c7a",
]


# Atlas color palette
atlas_colors = dict(
    {
        "neural_progenitor_cell": "#e9c46a",
        "telencephalic_npc": "#b77f92",
        "dorsal_npc": "#D19C2C",
        "dorsal_org": "#ffe082",
        "hippocampal_npc": "#F69B97",
        "ventral_npc": "#ce93d8",
        "ventral_org": "#9575cd",
        "dien_mesen_npc": "#80deea",
        "diencephalic_npc": "#29b6f6",
        "hypothalamic_npc": "#8EC7EC",
        "thalamic_npc": "#29D5F6",
        "mesencephalic_npc": "#26a69a",
        "retinal_npc": "#ffccbc",
        "rhombencephalic_npc": "#a5d6a7",
        "cerebellar_npc": "#95CB6E",
        "pons_npc": "#8BB599",
        "medullary_npc": "#BCD07E",
        "intermediate_progenitor": "#e76f51",
        "neuron": "#264653",
        "excitatory_neuron": "#A53259",
        "telencephalic_excitatory_neuron": "#b44a6e",
        "deeper_layer_cortical_excitatory_neuron": "#ec407a",
        "upper_layer_cortical_excitatory_neuron": "#8B1721",
        "hippocampal_excitatory_neuron": "#C35EA4",
        "non_telencephalic_excitatory_neuron": "#f9a825",
        "diencephalic_excitatory_neuron": "#0277bd",
        "thalamic_excitatory_neuron": "#2485BF",
        "hypothalamic_excitatory_neuron": "#134BBC",
        "mesencephalic_excitatory_neuron": "#00897b",
        "rhombencephalic_excitatory_neuron": "#558b2f",
        "cerebellar_excitatory_neuron": "#318C42",
        "pons_excitatory_neuron": "#4E703F",
        "medullary_excitatory_neuron": "#898B2F",
        "inhibitory_neuron": "#3A6890",
        "telencephalic_inhibitory_neuron": "#039be5",
        "ventral_inhibitory_neuron": "#76448a",
        "mge_inhibitory_neuron": "#8e44ad",
        "lge_inhibitory_neuron": "#5c6bc0",
        "cge_inhibitory_neuron": "#2B254B",
        "non_telencephalic_inhibitory_neuron": "#1B8466",
        "diencephalic_inhibitory_neuron": "#2196f3",
        "thalamic_inhibitory_neuron": "#5889A8",
        "hypothalamic_inhibitory_neuron": "#5A6EA8",
        "mesencephalic_inhibitory_neuron": "#00acc1",
        "rhombencephalic_inhibitory_neuron": "#4caf50",
        "cerebellar_inhibitory_neuron": "#699771",
        "pons_inhibitory_neuron": "#2F5D35",
        "medullary_inhibitory_neuron": "#C6B161",
        "choroid_plexus_epithelium": "#808b96",
        "astrocyte": "#DA2E7C",
        "oligodendrocyte_lineage": "#00bf99",
        "oligodendrocyte_precursor_cell": "#00FFCC",
        "mature_oligodendrocyte": "#00604d",
        "microglia": "#b2ff00",
        "vascular_endothelial_cell": "#F20000",
        "mesenchymal_cell": "#a1887f",
        "neural_crest": "#dce775",
        "pns_neurons": "#9e9d24",
    }
)



import anndata
import numpy as np
import scanpy as sc
from scipy import sparse


def get_matched_transcriptome(
    adata,
    adata_ref,
    wknn,
    rescale_factor=1,
):
    expr_ref = adata_ref.X.copy()
    if rescale_factor != 1:
        expr_ref = ((expr_ref.expm1()) * rescale_factor).log1p()

    normmat = sparse.diags(1 / np.array(wknn.sum(axis=1)).flatten())
    softmax_wknn = normmat.dot(wknn)
    expr_bg = softmax_wknn @ expr_ref

    adata_bg = anndata.AnnData(
        expr_bg, obs=adata.obs.copy(), var=adata_ref.var.copy(), obsm=adata.obsm.copy()
    )
    return adata_bg



from typing import Union, Optional
import anndata as ad
import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests

from .tests import ancova_group_test, f_nonzero_test, anova


def test_de(
    adata: ad.AnnData,
    group: Union[str, pd.Series],
    covar: Union[str, pd.DataFrame],
    num_threads: int = 1,
    return_coef_group: str = None,
    var_names: list = None,
    adjust_method: str = "holm",
) -> pd.DataFrame:
    """
    Test for differential expression using ANOVA

    Args:
        adata: AnnData object
        group: str or pd.Series
            The group labels
        covar: str or pd.DataFrame
            The covariates
        num_threads: int
            The number of threads to use
        return_coef_group: str
            The group to return coefficients for
        var_names: list
            The variable names to test
        adjust_method: str
            The method to adjust p-values. See https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html

    Returns:
        A `pd.DataFrame` with the differential expression results
    """
    if var_names is None:
        var_names = adata.var_names

    expr_mat = adata[:, var_names].X

    if isinstance(covar, (str, list, tuple)):
        covar = adata.obs[covar]
    elif isinstance(covar, pd.Series):
        covar = covar.values

    if isinstance(group, str):
        group = adata.obs[group]
    elif isinstance(group, pd.Series):
        group = group.values

    results = ancova_group_test(
        expr_mat,
        group,
        covar=covar,
        num_threads=num_threads,
        return_coef_group=return_coef_group,
        var_names=var_names,
    )

    results["padj"] = multipletests(results["p_Resi"], method=adjust_method)[1]
    results["pval"] = results["p_Resi"]
    return results


def test_de_paired(
    query_adata: ad.AnnData,
    matched_adata: ad.AnnData,
    covar: Optional[Union[str, pd.DataFrame]] = None,
    num_threads: int = 1,
    var_names: list = None,
    adjust_method: str = "holm",
) -> pd.DataFrame:
    """
    Test for differential expression between query data and matches reference cells using an F-test.

    Args:
        query_adata: AnnData object
            The query data
        matched_adata: AnnData object
            The matched reference data
        covar: str or pd.DataFrame
            The covariates
        num_threads: int
            The number of threads to use
        var_names: list
            The variable names to test
        adjust_method: str
            The method to adjust p-values. See https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html

    Returns:
        A `pd.DataFrame` with the differential expression results
    """
    if var_names is None:
        var_names = query_adata.var_names
    else:
        var_names = np.intersect1d(var_names, query_adata.var_names)

    var_names = np.intersect1d(var_names, matched_adata.var_names)

    if isinstance(covar, (str, list, tuple)):
        covar = query_adata.obs[covar]
    elif isinstance(covar, pd.Series):
        covar = covar.values

    expr_0 = query_adata[:, var_names].X - matched_adata[:, var_names].X

    results = f_nonzero_test(
        expr_0,
        covar=covar,
        num_threads=num_threads,
        var_names=var_names,
    )

    results["padj"] = multipletests(results["pval"], method=adjust_method)[1]
    return results



import warnings
import numpy as np
import pandas as pd
from scipy import sparse
import anndata as ad


def prepare_features(query_adata, ref_model):
    """
    Prepare the features of the query dataset to match the reference dataset
    """
    ref_features = ref_model.adata.var_names
    query_features = query_adata.var_names
    missing_features = np.setdiff1d(ref_features, query_features)

    if len(missing_features) == 0:
        return query_adata

    print(
        f"Warning: Query dataset is missing {len(missing_features)} features from the reference dataset. Adding missing features as zero-filled columns."
    )

    new_adata = query_adata.copy()
    new_adata = new_adata[:, np.isin(query_features, ref_features)]

    extra_X = sparse.csr_matrix(
        (new_adata.shape[0], len(missing_features)), dtype=new_adata.X.dtype
    )
    new_X = sparse.hstack((new_adata.X, extra_X))

    new_layers = {}
    for layer in new_adata.layers.keys():
        new_layers[layer] = sparse.hstack((new_adata.layers[layer], extra_X))

    new_var = pd.DataFrame(
        {"symbol": list(new_adata.var_names) + list(missing_features)}
    ).set_axis(list(new_adata.var_names) + list(missing_features))

    new_adata = ad.AnnData(
        X=new_X,
        obs=new_adata.obs.copy(),
        var=new_var,
        obsm=new_adata.obsm.copy(),
        layers=new_layers,
    )
    new_adata = new_adata[:, ref_features].copy()

    return new_adata



from pathlib import Path
from setuptools import setup, find_packages

author = "Jonas Simon Fleck, Zhisong He, Leander Dony"
author_email = "jonas.simon.fleck@gmail.com, zhisong.he@bsse.ethz.ch, leander.dony@helmholtz-munich.de"
description = "Human Neural Organoid Cell Atlas Toolbox"

long_description = Path("README.md").read_text("utf-8")
requirements = [
    l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
]

setup(
    name="hnoca",
    version="0.1.1",
    author=author,
    author_email=author_email,
    description=description,
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=requirements,
    python_requires=">=3.9",
)



from typing import Optional
import anndata as ad
import pandas as pd

import jax
from jax import numpy as jnp

from functools import partial

from .utils import get_expr, matrix_to_long_df
from .auroc import auc_expr


def find_markers(
    adata: ad.AnnData,
    group_name: str,
    features: Optional[list] = None,
    layer: Optional[str] = None,
) -> pd.DataFrame:
    """
    Find markers for each cluster.

    Args:
        adata: AnnData object
        group_name: Name of the column in adata.obs that contains the cluster labels
        features: List of features to use for marker identification
        layer: Layer in adata to use for expression

    Returns:
        A `pd.DataFrame` with AUROC and detection ratio for each gene and cluster.
    """
    # Get expression matrix
    expr, features = get_expr(adata, features=features, layer=layer)
    # Compute AUROC and fraction nonzero for marker features
    metrics = auc_expr(
        adata,
        group_name,
        features=features,
        compute_frac_nonzero_out=True,
        apply_fun="numpy",
    )
    # Make dataframes and join
    expr_df = matrix_to_long_df(
        metrics["auroc"], features=metrics["features"], groups=metrics["groups"]
    ).rename(columns={"value": "auroc"})

    auc_df = matrix_to_long_df(
        metrics["frac_nonzero"], features=metrics["features"], groups=metrics["groups"]
    ).rename(columns={"value": "frac_nonzero"})

    out_expr_df = matrix_to_long_df(
        metrics["frac_nonzero_out"],
        features=metrics["features"],
        groups=metrics["groups"],
    ).rename(columns={"value": "frac_nonzero_out"})

    marker_df = expr_df.merge(auc_df, on=["group", "feature"]).merge(
        out_expr_df, on=["group", "feature"]
    )
    # Compute detection ratio
    marker_df["detection_ratio"] = (
        marker_df["frac_nonzero"] / marker_df["frac_nonzero_out"]
    )

    return marker_df.sort_values(["group", "auroc"], ascending=False)



import os
import sys

import cloudpickle

from typing import Literal, Optional, Union

import anndata as ad
import numpy as np
import pandas as pd

import scvi
from scvi.model.utils import mde

import scarches
import scanpy as sc

from .wknn import get_wknn, transfer_labels, estimate_presence_score
from .matching import get_matched_transcriptome
from .utils import prepare_features


class AtlasMapper:
    """
    A class to map a query dataset to a reference dataset using scPoli, scVI or scANVI models.
    """

    def __init__(
        self,
        ref_model: Union[
            scvi.model.SCANVI, scvi.model.SCVI, scarches.models.scpoli.scPoli
        ],
    ):
        """
        Initialize the AtlasMapper object

        Args:
            ref_model: The reference model to map the query dataset to.
        """
        self.model_type = self._check_model_type(ref_model)
        self.ref_model = ref_model
        self.ref_adata = ref_model.adata
        self.query_model = None
        self.ref_trans_prob = None
        self.wknn = None

    def map_query(
        self,
        query_adata: ad.AnnData,
        retrain: Literal["partial", "full", "none"] = "partial",
        labeled_indices: Optional[np.ndarray] = None,
        **kwargs,
    ):
        """
        Map a query dataset to the reference dataset

        Args:
            query_adata: The query dataset to map to the reference dataset
            retrain: Whether to retrain the query model.

                * `"partial"` will retrain the weights of the new batch key
                * `"full"` will retrain the entire model
                * `"none"` will use the reference model without retraining

            labeled_indices: The indices of labeled cells in the query dataset. This is only used for scPoli models.
            **kwargs: Additional keyword arguments to pass to the training function
        """
        # Prepare the features of the query dataset
        query_adata = prepare_features(query_adata, self.ref_model)

        if retrain in ["partial", "full"]:
            if self.model_type == "scanvi":
                self._train_scanvi(query_adata, retrain, **kwargs)
            if self.model_type == "scvi":
                self._train_scvi(query_adata, retrain, **kwargs)
            if self.model_type == "scpoli":
                self._train_scpoli(query_adata, retrain, labeled_indices, **kwargs)
            self.query_adata = self.query_model.adata
        else:
            self.query_model = self.ref_model
            self.query_adata = query_adata

    def _train_scanvi(self, query_adata, retrain="partial", **kwargs):
        """
        Train a new scanvi model on the query data
        """
        unfrozen = retrain == "full"
        scvi.model.SCANVI.prepare_query_anndata(query_adata, self.ref_model)
        vae_q = scvi.model.SCANVI.load_query_data(
            query_adata, self.ref_model, unfrozen=unfrozen
        )
        vae_q.train(**kwargs)

        self.query_model = vae_q

    def _train_scvi(self, query_adata, retrain="partial", **kwargs):
        """
        Train a new scvi model on the query data
        """
        unfrozen = retrain == "full"
        scvi.model.SCVI.prepare_query_anndata(query_adata, self.ref_model)
        vae_q = scvi.model.SCVI.load_query_data(
            query_adata, self.ref_model, unfrozen=unfrozen
        )
        vae_q.train(**kwargs)

        self.query_model = vae_q

    def _train_scpoli(
        self, query_adata, retrain="partial", labeled_indices=None, **kwargs
    ):
        """
        Train a new scpoli model on the query data
        """
        freeze = retrain != "full"
        labeled_indices = [] if labeled_indices is None else labeled_indices

        # Set cell type to unknown if not present
        missing_cell_types = np.setdiff1d(
            np.array(self.ref_model.cell_type_keys_), query_adata.obs.columns
        )
        if len(missing_cell_types) > 0:
            query_adata.obs[missing_cell_types] = "Unknown"

        vae_q = scarches.models.scPoli.load_query_data(
            query_adata,
            reference_model=self.ref_model,
            unknown_ct_names=["Unknown"],
            labeled_indices=labeled_indices,
        )

        vae_q.train(**kwargs)

        self.query_model = vae_q

    def _check_model_type(self, model):
        if isinstance(model, scvi.model._scanvi.SCANVI):
            return "scanvi"
        elif isinstance(model, scvi.model._scvi.SCVI):
            return "scvi"
        elif isinstance(model, scarches.models.scpoli.scPoli):
            return "scpoli"
        else:
            raise RuntimeError("This VAE model is currently not supported")

    def _get_latent(self, model, adata, key=None, **kwargs):
        if key is not None and key in adata.obsm.keys():
            return adata.obsm[key]
        if self.model_type in ["scanvi", "scanvi"]:
            return model.get_latent_representation(adata, **kwargs)
        if self.model_type == "scpoli":
            return model.get_latent(adata, **kwargs)

    def get_latent_representation(self, adata=None, **kwargs):
        """
        Get the latent representation of the query and reference datasets

        Args:
            adata: The query or reference dataset
            **kwargs: Additional keyword arguments to pass to the get_latent function

        Returns:
            A tuple with the latent representation of the query and reference datasets
        """
        adata = self.query_adata if adata is None else adata
        model = self.query_model if self.query_model is not None else self.ref_model
        adata = prepare_features(adata, model)
        latent = self._get_latent(model, adata, **kwargs)
        return latent

    def compute_wknn(
        self,
        ref_adata: ad.AnnData = None,
        k: int = 100,
        query2ref: bool = True,
        ref2query: bool = False,
        ref_rep_key: str = "X_latent",
        query_rep_key: str = "X_latent",
        weighting_scheme: Literal[
            "n", "top_n", "jaccard", "jaccard_square", "gaussian", "dist"
        ] = "jaccard_square",
        top_n: Optional[int] = None,
    ):
        """
        Compute the weighted k-nearest neighbors graph between the reference and query datasets


        Args:
            k: Number of neighbors per cell
            query2ref: Consider query-to-ref neighbors
            ref2query: Consider ref-to-query neighbors
            weighting_scheme: How to weight edges in the ref-query neighbor graph
            top_n: The number of top neighbors to consider
        """
        
        self.ref_adata = ref_adata if ref_adata is not None else self.ref_adata
        ref_latent = self._get_latent(self.query_model, self.ref_adata, key=ref_rep_key)
        query_latent = self._get_latent(
            self.query_model, self.query_adata, key=query_rep_key
        )

        wknn = get_wknn(
            ref=ref_latent,
            query=query_latent,
            k=k,
            query2ref=query2ref,
            ref2query=ref2query,
            weighting_scheme=weighting_scheme,
            top_n=top_n,
        )

        self.wknn = wknn
        self.ref_adata.obsm["X_latent"] = ref_latent
        self.query_adata.obsm["X_latent"] = query_latent

    def get_presence_scores(
        self,
        split_by: str = None,
        random_walk: bool = True,
        alpha: float = 0.1,
        n_rounds: int = 100,
        log: bool = True,
    ) -> dict:
        """
        Estimate the presence score of the query dataset

        Args:
            split_by: The column in the query dataset to split by
            random_walk: Whether to use random walk to estimate presence score
            alpha: The heat diffusion parameter for the random walk
            n_rounds: The number of rounds for the random walk
            log: Whether to log the presence score

        Returns:
            A dictionary with the presence scores
        """

        if self.wknn is None:
            raise AttributeError("WKNN needs to be computed first.")

        scores = estimate_presence_score(
            self.ref_adata,
            self.query_adata,
            wknn=self.wknn,
            use_rep_ref_wknn="X_latent",
            use_rep_query_wknn="X_latent",
            ref_trans_prop=self.ref_trans_prob,
            split_by=split_by,
            alpha_random_walk=alpha,
            num_rounds_random_walk=n_rounds,
            log=log,
        )

        self.ref_trans_prob = scores["ref_trans_prop"]
        return scores

    def transfer_labels(self, label_key: str) -> dict:
        """
        Transfer labels from the reference dataset to the query dataset

        Args:
            label_key: str
                The column in the reference dataset to transfer

        Returns:
            A dictionary with the transfer scores
        """

        if self.wknn is None:
            raise AttributeError("WKNN needs to be computed first.")

        scores = transfer_labels(
            self.ref_adata,
            self.query_adata,
            self.wknn,
            label_key=label_key,
        )

        return scores

    def get_matched_expression(self, rescale_factor: int = 1) -> ad.AnnData:
        """
        Get the expression of reference cells matched to query cells. This can be used for quantitative comparisons like DE analysis.

        Args:
            rescale_factor: str
                Factor to rescale the log-normalized counts

        Returns:
            An AnnData object with the matched expression
        """

        if self.wknn is None:
            raise AttributeError("WKNN needs to be computed first.")

        matched_adata = get_matched_transcriptome(
            self.query_adata,
            self.ref_adata,
            self.wknn,
            rescale_factor=rescale_factor,
        )
        self.matched_adata = matched_adata
        return matched_adata

    def save(self, output_dir: str):
        """
        Save the mapper object to disk

        Args:
            output_dir: str
                The directory to save the mapper object
        """
        os.makedirs(output_dir, exist_ok=True)
        with open(os.path.join(output_dir, "mapper.pkl"), "wb") as f:
            cloudpickle.dump(self, f)

    @classmethod
    def load(cls, input_dir: str):
        """
        Load the mapper object from disk

        Args:
            input_dir: str
                The directory to load the mapper object
        """
        with open(os.path.join(input_dir, "mapper.pkl"), "rb") as f:
            mapper = cloudpickle.load(f)
        return mapper
